'''A robots.txt cache.'''

from functools import partial
import threading
import time

from cachetools import LRUCache

from .policy import DefaultObjectPolicy, ReraiseExceptionPolicy
from ..robots import Robots, AllowNone, Agent
from .. import logger


class ExpiringObject(object):
    '''Wrap an object that expires over time.'''

    def __init__(self, factory):
        self.factory = factory
        self.lock = threading.Lock()
        self.obj = None
        self.expires = 0
        self.exception = None

    def get(self):
        '''Get the wrapped object.'''
        if (self.obj is None) or (time.time() >= self.expires):
            with self.lock:
                self.expires, self.obj = self.factory()
                if isinstance(self.obj, BaseException):
                    self.exception = self.obj
                else:
                    self.exception = None

        if self.exception:
            raise self.exception
        else:
            return self.obj


class BaseCache(object):
    '''A base cache class.'''

    DEFAULT_CACHE_POLICY = ReraiseExceptionPolicy(ttl=600)
    DEFAULT_TTL_POLICY = Robots.DEFAULT_TTL_POLICY

    def __init__(self, capacity, cache_policy=None, ttl_policy=None, *args, **kwargs):
        self.cache_policy = cache_policy or self.DEFAULT_CACHE_POLICY
        self.ttl_policy = ttl_policy or self.DEFAULT_TTL_POLICY
        self.cache = LRUCache(maxsize=capacity)
        self.args = args
        self.kwargs = kwargs

    def get(self, url):
        '''Get the entity that corresponds to URL.'''
        robots_url = Robots.robots_url(url)
        if robots_url not in self.cache:
            self.cache[robots_url] = ExpiringObject(partial(self.factory, robots_url))
        return self.cache[robots_url].get()

    def factory(self, url):
        '''
        Return (expiration, obj) corresponding to provided url, exercising the
        cache_policy as necessary.
        '''
        try:
            return self.fetch(url)
        except BaseException as exc:
            logger.exception('Reppy cache fetch error on %s' % url)
            return self.cache_policy.exception(url, exc)

    def fetch(self, url):
        '''Return (expiration, obj) corresponding to provided url.'''
        raise NotImplementedError('BaseCache does not implement fetch.')


class RobotsCache(BaseCache):
    '''A cache of Robots objects.'''

    DEFAULT_CACHE_POLICY = DefaultObjectPolicy(ttl=600, factory=AllowNone)

    def allowed(self, url, agent):
        '''Return true if the provided URL is allowed to agent.'''
        return self.get(url).allowed(url, agent)

    def fetch(self, url):
        '''Return (expiration, Robots) for the robots.txt at the provided URL.'''
        robots = Robots.fetch(
            url, ttl_policy=self.ttl_policy, *self.args, **self.kwargs)
        return (robots.expires, robots)


class AgentCache(BaseCache):
    '''A cache of Agent objects.'''

    DEFAULT_CACHE_POLICY = DefaultObjectPolicy(
        ttl=600, factory=lambda url: Agent().disallow('/'))

    def __init__(self, agent, *args, **kwargs):
        BaseCache.__init__(self, *args, **kwargs)
        self.agent = agent

    def allowed(self, url):
        '''Return true if the provided URL is allowed to self.agent.'''
        return self.get(url).allowed(url)

    def fetch(self, url):
        '''Return (expiration, Agent) for the robots.txt at the provided URL.'''
        robots = Robots.fetch(
            url, ttl_policy=self.ttl_policy, *self.args, **self.kwargs)
        return (robots.expires, robots.agent(self.agent))



'''Policies for caching.'''

import time


class CachePolicyBase(object):
    '''Policy for caching.'''

    def exception(self, url, exception):
        '''What to return when there's an exception.'''
        raise NotImplementedError('CachePolicyBase does not implement exception.')


class DefaultObjectPolicy(object):
    '''Return a default object on exception.'''

    def __init__(self, ttl, factory):
        self.ttl = ttl
        self.factory = factory

    def exception(self, url, exception):
        '''What to return when there's an exception.'''
        return (time.time() + self.ttl, self.factory(url))


class ReraiseExceptionPolicy(object):
    '''Reraise the exception.'''

    def __init__(self, ttl):
        self.ttl = ttl

    def exception(self, url, exception):
        '''What to return when there's an exception.'''
        return (time.time() + self.ttl, exception)



'''Robots.txt parsing.'''

import logging

logger = logging.getLogger('reppy')
handler = logging.StreamHandler()
formatter = logging.Formatter(
    ' | '.join([
        '[%(asctime)s]',
        'PID %(process)d',
        '%(levelname)s in %(module)s:%(funcName)s@%(lineno)s => %(message)s'
    ]))
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)
logger.setLevel(logging.ERROR)

from .robots import Robots, Agent



#! /usr/bin/env python
#
# Copyright (c) 2011 SEOmoz
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''All of our exceptions'''


class ReppyException(Exception):
    '''Any internal exception'''
    pass

class ContentTooLong(ReppyException):
    '''Robots.txt content is too long.'''
    pass

class ServerError(ReppyException):
    '''When the remote server returns an error'''
    pass

class SSLException(ReppyException):
    '''An SSL error.'''
    pass

class ConnectionException(ReppyException):
    '''A connection error exception.'''
    pass

class MalformedUrl(ReppyException):
    '''An exception for a missing or invalid url or schema.'''
    pass

class ExcessiveRedirects(ReppyException):
    '''A TooManyRedirects error.'''
    pass

class ReadTimeout(ReppyException):
    '''A ReadTimeout error from the HTTP library.'''
    pass

class BadStatusCode(ReppyException):
    '''An exception for 5xx status codes.'''
    pass



'''Policies for setting the TTL on Robots objects.'''

import time

from . import logger
from .util import parse_date


class TTLPolicyBase(object):
    '''Policy for setting the TTL on Robots objects.'''

    def ttl(self, response):
        '''Get the caching TTL for a response.'''
        raise NotImplementedError('TTLPolicyBase does not implement ttl.')

    def expires(self, response):
        '''Determine when a response should expire.'''
        return time.time() + self.ttl(response)


class HeaderWithDefaultPolicy(TTLPolicyBase):
    '''TTL is based on headers, but falls back to a default, clamps to a minimum.'''

    def __init__(self, default, minimum):
        self.default = default
        self.minimum = minimum

    def ttl(self, response):
        '''Get the ttl from headers.'''
        # If max-age is specified in Cache-Control, use it and ignore any
        # Expires header, as per RFC2616 Sec. 13.2.4.
        cache_control = response.headers.get('cache-control')
        if cache_control is not None:
            for directive in cache_control.split(','):
                name, _, value = directive.lower().partition('=')
                name = name.strip()
                if name in ('no-store', 'must-revalidate', 'no-cache'):
                    return max(self.minimum, 0)
                elif name in ('s-maxage', 'max-age'):
                    try:
                        return max(self.minimum, int(value.strip()))
                    except ValueError:
                        logger.warn(
                            'Could not parse %s=%s', name, value, exc_info=1)

        # Check the Expires header
        expires = response.headers.get('expires')
        if expires is not None:
            # Evaluate the expiration relative to the server-provided date
            date = response.headers.get('date')
            if date is not None:
                try:
                    date = parse_date(date)
                except ValueError:
                    logger.warn(
                        'Could not parse date string %s', date, exc_info=1)
                    date = time.time()
            else:
                date = time.time()

            try:
                return max(self.minimum, parse_date(expires) - date)
            except ValueError:
                logger.warn(
                    'Could not parse date string %s', expires, exc_info=1)

        return self.default



'''Utility functions.'''

import email


def parse_date(string):
    '''Return a timestamp for the provided datestring, described by RFC 7231.'''
    parsed = email.utils.parsedate_tz(string)
    if parsed is None:
        raise ValueError("Invalid time.")
    parsed = list(parsed)
    # Default time zone is GMT/UTC
    parsed[9] = 0 if parsed[9] is None else parsed[9]
    return email.utils.mktime_tz(parsed)



# Cython declarations

from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp cimport bool

cdef extern from "rep-cpp/include/directive.h" namespace "Rep":
    cpdef cppclass CppDirective "Rep::Directive":
        ctypedef size_t priority_t

        CppDirective(const string& line, bool allowed)
        CppDirective(const CppDirective& rhs)
        priority_t priority() const
        bool match(const string& path) const
        bool allowed() const
        string str() const

cdef extern from "rep-cpp/include/agent.h" namespace "Rep":
    cpdef cppclass CppAgent "Rep::Agent":
        ctypedef float delay_t

        CppAgent()
        CppAgent(const string& host)
        CppAgent& allow(const string& query)
        CppAgent& disallow(const string& query)
        CppAgent& delay(delay_t delay)
        delay_t delay() const
        const vector[CppDirective]& directives() const
        bool allowed(const string& path) const
        string str() const

cdef extern from "rep-cpp/include/robots.h" namespace "Rep":
    cpdef cppclass CppRobots "Rep::Robots":
        CppRobots(const string& content) except +ValueError
        CppRobots(const string& content, const string& base_url) except +ValueError
        const vector[string]& sitemaps() const
        CppAgent& agent(const string& name) const
        bool allowed(const string& path, const string& name) const
        string str() const
        @staticmethod
        string robotsUrl(const string& url) except +ValueError



#!/usr/bin/env python

# Copyright (c) 2011-2017 SEOmoz, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from setuptools import setup
from setuptools.extension import Extension

ext_files = [
    'reppy/rep-cpp/src/agent.cpp',
    'reppy/rep-cpp/src/directive.cpp',
    'reppy/rep-cpp/src/robots.cpp',
    'reppy/rep-cpp/deps/url-cpp/src/url.cpp',
    'reppy/rep-cpp/deps/url-cpp/src/utf8.cpp',
    'reppy/rep-cpp/deps/url-cpp/src/punycode.cpp',
    'reppy/rep-cpp/deps/url-cpp/src/psl.cpp'
]

kwargs = {}

try:
    from Cython.Distutils import build_ext
    print('Building from Cython')
    ext_files.append('reppy/robots.pyx')
    kwargs['cmdclass'] = {'build_ext': build_ext}
except ImportError:
    print('Building from C++')
    ext_files.append('reppy/robots.cpp')

ext_modules = [
    Extension(
        'reppy.robots', ext_files,
        language='c++',
        extra_compile_args=['-std=c++11'],
        include_dirs=[
            'reppy/rep-cpp/include',
            'reppy/rep-cpp/deps/url-cpp/include'])
]

setup(
    name='reppy',
    version='0.4.13',
    description='Replacement robots.txt Parser',
    long_description='''Replaces the built-in robotsparser with a
RFC-conformant implementation that supports modern robots.txt constructs like
Sitemaps, Allow, and Crawl-delay. Main features:

- Memoization of fetched robots.txt
- Expiration taken from the `Expires` header
- Batch queries
- Configurable user agent for fetching robots.txt
- Automatic refetching based on expiration
''',
    maintainer='Lindsey Reno',
    maintainer_email='lindsey@moz.com',
    url='http://github.com/seomoz/reppy',
    license='MIT',
    platforms='Posix; MacOS X',
    ext_modules=ext_modules,
    packages=[
        'reppy',
        'reppy.cache'
    ],
    package_dir={
        'reppy': 'reppy',
        'reppy.cache': 'reppy/cache'
    },
    install_requires=[
        'cachetools',
        'python-dateutil>=1.5, !=2.0',
        'requests',
        'six'
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    **kwargs
)



# cython: linetrace=True
# distutils: define_macros=CYTHON_TRACE=1

from contextlib import closing
import time

import requests
from requests.exceptions import (
    SSLError,
    ConnectionError,
    URLRequired,
    MissingSchema,
    InvalidSchema,
    InvalidURL,
    TooManyRedirects,
    ReadTimeout)
import six

from .ttl import HeaderWithDefaultPolicy
from . import util, logger, exceptions

cdef as_bytes(value):
    if isinstance(value, bytes):
        return value
    return value.encode('utf-8')

# For contexts which require a 'str' type, convert bytes to unicode if needed
# (i.e., Python 3). Note: could raise UnicodeDecodeError in Python 3 if input
# is invalid UTF-8
cdef as_string(value):
    if six.PY3:
        if isinstance(value, bytes):
            return value.decode('utf-8')
    return value


def FromRobotsMethod(cls, Robots robots, const string& name):
    '''Construct an Agent from a CppAgent.'''
    agent = Agent()
    # This is somewhat inefficient due to the copying, but it is
    # required to be copied because we often toss the containing
    # Robots object as a temporary thus we'd leave the underlying
    # Agent object dangling without a full copy.
    agent.agent = robots.robots.agent(name)
    return agent

cdef class Agent:
    '''Wrapper around rep-cpp's Rep::Agent class.'''

    cdef CppAgent agent

    from_robots = classmethod(FromRobotsMethod)

    def __str__(self):
        return as_string(self.agent.str())

    def __len__(self):
        return self.agent.directives().size()

    @property
    def delay(self):
        '''The delay associated with this agent.'''
        cdef float value = self.agent.delay()
        if value > 0:
            return value
        return None

    def allow(self, path):
        '''Allow the provided path.'''
        self.agent.allow(as_bytes(path))
        return self

    def disallow(self, path):
        '''Disallow the provided path.'''
        self.agent.disallow(as_bytes(path))
        return self

    def allowed(self, path):
        '''Is the provided URL allowed?'''
        return self.agent.allowed(as_bytes(path))


def ParseMethod(cls, url, content, expires=None):
    '''Parse a robots.txt file.'''
    return cls(url, as_bytes(content), expires)

def FetchMethod(cls, url, ttl_policy=None, max_size=1048576, *args, **kwargs):
    '''Get the robots.txt at the provided URL.'''
    after_response_hook = kwargs.pop('after_response_hook', None)
    after_parse_hook = kwargs.pop('after_parse_hook', None)
    def wrap_exception(etype, cause):
        wrapped = etype(cause)
        wrapped.url = url
        if after_response_hook is not None:
            after_response_hook(wrapped)
        raise wrapped
    try:
        # Limit the size of the request
        kwargs['stream'] = True
        with closing(requests.get(url, *args, **kwargs)) as res:
            content = res.raw.read(amt=max_size, decode_content=True)
            # Try to read an additional byte, to see if the response is too big
            if res.raw.read(amt=1, decode_content=True):
                raise exceptions.ContentTooLong(
                    'Content larger than %s bytes' % max_size)

            if after_response_hook is not None:
                after_response_hook(res)

            # Get the TTL policy's ruling on the ttl
            expires = (ttl_policy or cls.DEFAULT_TTL_POLICY).expires(res)

            if res.status_code == 200:
                robots = cls.parse(url, content, expires)
                if after_parse_hook is not None:
                    after_parse_hook(robots)
                return robots
            elif res.status_code in (401, 403):
                return AllowNone(url, expires)
            elif res.status_code >= 400 and res.status_code < 500:
                return AllowAll(url, expires)
            else:
                raise exceptions.BadStatusCode(
                    'Got %i for %s' % (res.status_code, url), res.status_code)
    except SSLError as exc:
        wrap_exception(exceptions.SSLException, exc)
    except ConnectionError as exc:
        wrap_exception(exceptions.ConnectionException, exc)
    except (URLRequired, MissingSchema, InvalidSchema, InvalidURL) as exc:
        wrap_exception(exceptions.MalformedUrl, exc)
    except TooManyRedirects as exc:
        wrap_exception(exceptions.ExcessiveRedirects, exc)
    except ReadTimeout as exc:
        wrap_exception(exceptions.ReadTimeout, exc)

def RobotsUrlMethod(cls, url):
    '''Get the robots.txt URL that corresponds to the provided one.'''
    return as_string(CppRobots.robotsUrl(as_bytes(url)))

cdef class Robots:
    '''Wrapper around rep-cpp's Rep::Robots class.'''

    # The default TTL policy is to cache for 3600 seconds or what's provided in the
    # headers, and a minimum of 600 seconds
    DEFAULT_TTL_POLICY = HeaderWithDefaultPolicy(default=3600, minimum=600)

    # Class methods
    parse = classmethod(ParseMethod)
    fetch = classmethod(FetchMethod)
    robots_url = classmethod(RobotsUrlMethod)

    # Data members
    cdef CppRobots* robots
    cdef object expires

    def __init__(self, url, const string& content, expires=None):
        self.robots = new CppRobots(content, as_bytes(url))
        self.expires = expires

    def __str__(self):
        # Note: this could raise a UnicodeDecodeError in Python 3 if the
        # robots.txt had invalid UTF-8
        return as_string(self.robots.str())

    def __dealloc__(self):
        del self.robots

    @property
    def sitemaps(self):
        '''Get all the sitemaps in this robots.txt.'''
        return map(as_string, self.robots.sitemaps())

    def allowed(self, path, name):
        '''Is the provided path allowed for the provided agent?'''
        return self.robots.allowed(as_bytes(path), as_bytes(name))

    def agent(self, name):
        '''Return the Agent that corresponds to name.

        Note modifications to the returned Agent will not be reflected
        in this Robots object because it is a *copy*, not the original
        Agent object.
        '''
        return Agent.from_robots(self, as_bytes(name))

    @property
    def expired(self):
        '''True if the current time is past its expiration.'''
        return time.time() > self.expires

    @property
    def expires(self):
        '''The expiration of this robots.txt.'''
        return self.expires

    @property
    def ttl(self):
        '''Remaining time for this response to be considered valid.'''
        return max(self.expires - time.time(), 0)


cdef class AllowNone(Robots):
    '''No requests are allowed.'''

    def __init__(self, url, expires=None):
        Robots.__init__(self, url, b'User-agent: *\nDisallow: /', expires)


cdef class AllowAll(Robots):
    '''All requests are allowed.'''

    def __init__(self, url, expires=None):
        Robots.__init__(self, url, b'', expires)
